@@ -1,51 +1,64 @@
-/* Copyright (C) 2009-2015, 2019 Free Software Foundation, Inc.
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 2009-2015,2018-2019,2023
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
 
-#include <alloca.h>
-#include <assert.h>
-
-#include <gmp.h>
-
-#include "libguile/_scm.h"
-#include "libguile/extensions.h"
-#include "libguile/bytevectors.h"
-#include "libguile/strings.h"
-#include "libguile/validate.h"
-#include "libguile/arrays.h"
-#include "libguile/array-handle.h"
-#include "libguile/uniform.h"
-#include "libguile/srfi-4.h"
-
+#include <limits.h>
 #include <byteswap.h>
+#include <intprops.h>
+#include <errno.h>
 #include <striconveh.h>
 #include <uniconv.h>
 #include <unistr.h>
+#include <string.h>
+#include <alloca.h>
+#include <assert.h>
 
-#ifdef HAVE_LIMITS_H
-# include <limits.h>
+#include "scm.h"
+
+#if SCM_ENABLE_MINI_GMP
+#include "mini-gmp.h"
+#else
+#include <gmp.h>
 #endif
 
-#include <string.h>
+#include "array-handle.h"
+#include "arrays.h"
+#include "boolean.h"
+#include "dynwind.h"
+#include "extensions.h"
+#include "generalized-vectors.h"
+#include "gsubr.h"
+#include "list.h"
+#include "numbers.h"
+#include "pairs.h"
+#include "ports.h"
+#include "srfi-4.h"
+#include "strings.h"
+#include "symbols.h"
+#include "uniform.h"
+#include "version.h"
+
+#include "bytevectors.h"
 
 
 
@@ -53,12 +66,12 @@
 
 /* Convenience macros.  These are used by the various templates (macros) that
    are parameterized by integer signedness.  */
-#define INT8_T_signed           scm_t_int8
-#define INT8_T_unsigned         scm_t_uint8
-#define INT16_T_signed          scm_t_int16
-#define INT16_T_unsigned        scm_t_uint16
-#define INT32_T_signed          scm_t_int32
-#define INT32_T_unsigned        scm_t_uint32
+#define INT8_T_signed           int8_t
+#define INT8_T_unsigned         uint8_t
+#define INT16_T_signed          int16_t
+#define INT16_T_unsigned        uint16_t
+#define INT32_T_signed          int32_t
+#define INT32_T_unsigned        uint32_t
 #define is_signed_int8(_x)      (((_x) >= -128L) && ((_x) <= 127L))
 #define is_unsigned_int8(_x)    ((_x) <= 255UL)
 #define is_signed_int16(_x)     (((_x) >= -32768L) && ((_x) <= 32767L))
@@ -313,6 +326,69 @@ scm_c_take_typed_bytevector (signed char *contents, size_t len,
   return ret;
 }
 
+SCM_DEFINE (scm_bytevector_slice, "bytevector-slice", 2, 1, 0,
+            (SCM bv, SCM offset, SCM size),
+            "Return the slice of @var{bv} starting at @var{offset} and counting\n"
+            "@var{size} bytes.  When @var{size} is omitted, the slice covers all\n"
+            "of @var{bv} starting from @var{offset}.  The returned slice shares\n"
+            "storage with @var{bv}: changes to the slice are visible in @var{bv}\n"
+            "and vice-versa.\n"
+            "\n"
+            "When @var{bv} is actually a SRFI-4 uniform vector, its element\n"
+            "type is preserved unless @var{offset} and @var{size} are not aligned\n"
+            "on its element type size.\n")
+#define FUNC_NAME s_scm_bytevector_slice
+{
+  SCM ret;
+  size_t c_offset, c_size;
+  scm_t_array_element_type element_type;
+
+  SCM_VALIDATE_BYTEVECTOR (1, bv);
+
+  c_offset = scm_to_size_t (offset);
+
+  if (SCM_UNBNDP (size))
+    {
+      if (c_offset < SCM_BYTEVECTOR_LENGTH (bv))
+        c_size = SCM_BYTEVECTOR_LENGTH (bv) - c_offset;
+      else
+        c_size = 0;
+    }
+  else
+    c_size = scm_to_size_t (size);
+
+  if (c_offset == 0 && c_size == SCM_BYTEVECTOR_LENGTH (bv))
+    return bv;
+
+  if (INT_ADD_OVERFLOW (c_offset, c_size)
+      || (c_offset + c_size > SCM_BYTEVECTOR_LENGTH (bv)))
+    scm_out_of_range (FUNC_NAME, offset);
+
+  /* Preserve the element type of BV, unless we're not slicing on type
+     boundaries.  */
+  element_type = SCM_BYTEVECTOR_ELEMENT_TYPE (bv);
+  if ((c_offset % SCM_BYTEVECTOR_TYPE_SIZE (bv) != 0)
+      || (c_size % SCM_BYTEVECTOR_TYPE_SIZE (bv) != 0))
+    element_type = SCM_ARRAY_ELEMENT_TYPE_VU8;
+  else
+    c_size /= (scm_i_array_element_type_sizes[element_type] / 8);
+
+  ret = make_bytevector_from_buffer (c_size,
+                                     SCM_BYTEVECTOR_CONTENTS (bv) + c_offset,
+                                     element_type);
+  if (!SCM_MUTABLE_BYTEVECTOR_P (bv))
+    {
+      /* Preserve the immutability property.  */
+      scm_t_bits flags = SCM_BYTEVECTOR_FLAGS (ret);
+      SCM_SET_BYTEVECTOR_FLAGS (ret, flags | SCM_F_BYTEVECTOR_IMMUTABLE);
+    }
+
+  SCM_BYTEVECTOR_SET_PARENT (ret, bv);
+
+  return ret;
+}
+#undef FUNC_NAME
+
 /* Shrink BV to C_NEW_LEN (which is assumed to be smaller than its current
    size) and return the new bytevector (possibly different from BV).  */
 SCM
@@ -372,17 +448,17 @@ scm_c_bytevector_length (SCM bv)
 }
 #undef FUNC_NAME
 
-scm_t_uint8
+uint8_t
 scm_c_bytevector_ref (SCM bv, size_t index)
 #define FUNC_NAME "scm_c_bytevector_ref"
 {
   size_t c_len;
-  const scm_t_uint8 *c_bv;
+  const uint8_t *c_bv;
 
   SCM_VALIDATE_BYTEVECTOR (1, bv);
 
   c_len = SCM_BYTEVECTOR_LENGTH (bv);
-  c_bv = (scm_t_uint8 *) SCM_BYTEVECTOR_CONTENTS (bv);
+  c_bv = (uint8_t *) SCM_BYTEVECTOR_CONTENTS (bv);
 
   if (SCM_UNLIKELY (index >= c_len))
     scm_out_of_range (FUNC_NAME, scm_from_size_t (index));
@@ -392,16 +468,16 @@ scm_c_bytevector_ref (SCM bv, size_t index)
 #undef FUNC_NAME
 
 void
-scm_c_bytevector_set_x (SCM bv, size_t index, scm_t_uint8 value)
+scm_c_bytevector_set_x (SCM bv, size_t index, uint8_t value)
 #define FUNC_NAME "scm_c_bytevector_set_x"
 {
   size_t c_len;
-  scm_t_uint8 *c_bv;
+  uint8_t *c_bv;
 
   SCM_VALIDATE_MUTABLE_BYTEVECTOR (1, bv);
 
   c_len = SCM_BYTEVECTOR_LENGTH (bv);
-  c_bv = (scm_t_uint8 *) SCM_BYTEVECTOR_CONTENTS (bv);
+  c_bv = (uint8_t *) SCM_BYTEVECTOR_CONTENTS (bv);
 
   if (SCM_UNLIKELY (index >= c_len))
     scm_out_of_range (FUNC_NAME, scm_from_size_t (index));
@@ -481,7 +557,7 @@ SCM_DEFINE (scm_make_bytevector, "make-bytevector", 1, 1, 0,
 {
   SCM bv;
   size_t c_len;
-  scm_t_uint8 c_fill = 0;
+  uint8_t c_fill = 0;
 
   SCM_VALIDATE_SIZE_COPY (1, len, c_len);
   if (!scm_is_eq (fill, SCM_UNDEFINED))
@@ -491,16 +567,16 @@ SCM_DEFINE (scm_make_bytevector, "make-bytevector", 1, 1, 0,
       value = scm_to_int (fill);
       if (SCM_UNLIKELY ((value < -128) || (value > 255)))
 	scm_out_of_range (FUNC_NAME, fill);
-      c_fill = (scm_t_uint8) value;
+      c_fill = (uint8_t) value;
     }
 
   bv = make_bytevector (c_len, SCM_ARRAY_ELEMENT_TYPE_VU8);
   if (!scm_is_eq (fill, SCM_UNDEFINED))
     {
       size_t i;
-      scm_t_uint8 *contents;
+      uint8_t *contents;
 
-      contents = (scm_t_uint8 *) SCM_BYTEVECTOR_CONTENTS (bv);
+      contents = (uint8_t *) SCM_BYTEVECTOR_CONTENTS (bv);
       for (i = 0; i < c_len; i++)
 	contents[i] = c_fill;
     }
@@ -550,39 +626,57 @@ SCM_DEFINE (scm_bytevector_eq_p, "bytevector=?", 2, 0, 0,
 }
 #undef FUNC_NAME
 
-SCM_DEFINE (scm_bytevector_fill_x, "bytevector-fill!", 2, 0, 0,
-	    (SCM bv, SCM fill),
-	    "Fill bytevector @var{bv} with @var{fill}, a byte.")
-#define FUNC_NAME s_scm_bytevector_fill_x
-{
-  size_t c_len, i;
-  scm_t_uint8 *c_bv, c_fill;
-  int value;
+static SCM scm_bytevector_fill_partial_x (SCM bv, SCM fill, SCM start, SCM end);
 
+SCM_DEFINE (scm_bytevector_fill_partial_x, "bytevector-fill!", 2, 2, 0,
+	    (SCM bv, SCM fill, SCM start, SCM end),
+	    "Fill positions [@var{start} ... @var{end}) of bytevector "
+            "@var{bv} with @var{fill}, a byte. @var{start} defaults to 0 "
+            "and @var{end} defaults to the length of @var{bv}. "
+            "The return value is unspecified.")
+#define FUNC_NAME s_scm_bytevector_fill_partial_x
+{
   SCM_VALIDATE_MUTABLE_BYTEVECTOR (1, bv);
 
-  value = scm_to_int (fill);
+  int value  = scm_to_int (fill);
   if (SCM_UNLIKELY ((value < -128) || (value > 255)))
     scm_out_of_range (FUNC_NAME, fill);
-  c_fill = (scm_t_uint8) value;
+  
+  size_t i = 0;
+  size_t c_end = SCM_BYTEVECTOR_LENGTH (bv);
+  uint8_t *c_bv = (uint8_t *) SCM_BYTEVECTOR_CONTENTS (bv);
 
-  c_len = SCM_BYTEVECTOR_LENGTH (bv);
-  c_bv = (scm_t_uint8 *) SCM_BYTEVECTOR_CONTENTS (bv);
+  if (!SCM_UNBNDP (start))
+    i = scm_to_unsigned_integer (start, 0, c_end);
+  if (!SCM_UNBNDP (end))
+    c_end = scm_to_unsigned_integer (end, i, c_end);
 
-  for (i = 0; i < c_len; i++)
-    c_bv[i] = c_fill;
+  memset (c_bv + i, value, c_end-i);
 
   return SCM_UNSPECIFIED;
 }
 #undef FUNC_NAME
 
+SCM
+scm_bytevector_fill_x (SCM bv, SCM fill)
+#define FUNC_NAME s_scm_bytevector_fill_x
+{
+  return scm_bytevector_fill_partial_x (bv, fill, SCM_UNDEFINED, SCM_UNDEFINED);
+}
+#undef FUNC_NAME
+
+
 SCM_DEFINE (scm_bytevector_copy_x, "bytevector-copy!", 5, 0, 0,
 	    (SCM source, SCM source_start, SCM target, SCM target_start,
 	     SCM len),
 	    "Copy @var{len} bytes from @var{source} into @var{target}, "
-	    "starting reading from @var{source_start} (a positive index "
-	    "within @var{source}) and start writing at "
-	    "@var{target_start}.")
+	    "reading from a block starting at @var{source_start} (a positive "
+            "index within @var{source}) and writing to a block starting at "
+	    "@var{target_start}.\n\n"
+            "It is permitted for the @var{source} and @var{target} regions to "
+            "overlap. In that case, copying takes place as if the source is "
+            "first copied into a temporary bytevector and then into the "
+            "destination. ")
 #define FUNC_NAME s_scm_bytevector_copy_x
 {
   size_t c_len, c_source_len, c_target_len;
@@ -730,12 +824,12 @@ SCM_DEFINE (scm_bytevector_to_u8_list, "bytevector->u8-list", 1, 0, 0,
 {
   SCM lst, pair;
   size_t c_len, i;
-  scm_t_uint8 *c_bv;
+  uint8_t *c_bv;
 
   SCM_VALIDATE_BYTEVECTOR (1, bv);
 
   c_len = SCM_BYTEVECTOR_LENGTH (bv);
-  c_bv = (scm_t_uint8 *) SCM_BYTEVECTOR_CONTENTS (bv);
+  c_bv = (uint8_t *) SCM_BYTEVECTOR_CONTENTS (bv);
 
   lst = scm_make_list (scm_from_size_t (c_len), SCM_UNSPECIFIED);
   for (i = 0, pair = lst;
@@ -756,12 +850,12 @@ SCM_DEFINE (scm_u8_list_to_bytevector, "u8-list->bytevector", 1, 0, 0,
 {
   SCM bv, item;
   size_t c_len, i;
-  scm_t_uint8 *c_bv;
+  uint8_t *c_bv;
 
   SCM_VALIDATE_LIST_COPYLEN (1, lst, c_len);
 
   bv = make_bytevector (c_len, SCM_ARRAY_ELEMENT_TYPE_VU8);
-  c_bv = (scm_t_uint8 *) SCM_BYTEVECTOR_CONTENTS (bv);
+  c_bv = (uint8_t *) SCM_BYTEVECTOR_CONTENTS (bv);
 
   for (i = 0; i < c_len; lst = SCM_CDR (lst), i++)
     {
@@ -773,7 +867,7 @@ SCM_DEFINE (scm_u8_list_to_bytevector, "u8-list->bytevector", 1, 0, 0,
 
 	  c_item = SCM_I_INUM (item);
 	  if (SCM_LIKELY ((c_item >= 0) && (c_item < 256)))
-	    c_bv[i] = (scm_t_uint8) c_item;
+	    c_bv[i] = (uint8_t) c_item;
 	  else
 	    goto type_error;
 	}
@@ -1610,13 +1704,13 @@ SCM_DEFINE (scm_bytevector_s64_native_set_x, "bytevector-s64-native-set!",
 union scm_ieee754_float
 {
   float f;
-  scm_t_uint32 i;
+  uint32_t i;
 };
 
 union scm_ieee754_double
 {
   double d;
-  scm_t_uint64 i;
+  uint64_t i;
 };
 
 
@@ -1921,7 +2015,7 @@ utf_encoding_name (char *name, size_t utf_width, SCM endianness)
                                                                         \
   SCM_VALIDATE_STRING (1, str);                                         \
   if (scm_is_eq (endianness, SCM_UNDEFINED))                            \
-    endianness = sym_big;                                           \
+    endianness = sym_big;                                               \
   else                                                                  \
     SCM_VALIDATE_SYMBOL (2, endianness);                                \
                                                                         \
@@ -1943,7 +2037,7 @@ utf_encoding_name (char *name, size_t utf_width, SCM endianness)
       const scm_t_wchar *wbuf = scm_i_string_wide_chars (str);          \
       c_utf = u32_conv_to_encoding (c_utf_name,                         \
                                     iconveh_question_mark,              \
-                                    (scm_t_uint32 *) wbuf,              \
+                                    (uint32_t *) wbuf,                  \
                                     c_strlen, NULL, NULL, &c_utf_len);  \
       if (SCM_UNLIKELY (c_utf == NULL))                                 \
         scm_syserror_msg (FUNC_NAME, "failed to convert string: ~A",    \
@@ -1967,12 +2061,12 @@ SCM_DEFINE (scm_string_to_utf8, "string->utf8",
 #define FUNC_NAME s_scm_string_to_utf8
 {
   SCM utf;
-  scm_t_uint8 *c_utf;
+  uint8_t *c_utf;
   size_t c_utf_len = 0;
 
   SCM_VALIDATE_STRING (1, str);
 
-  c_utf = (scm_t_uint8 *) scm_to_utf8_stringn (str, &c_utf_len);
+  c_utf = (uint8_t *) scm_to_utf8_stringn (str, &c_utf_len);
   utf = make_bytevector (c_utf_len, SCM_ARRAY_ELEMENT_TYPE_VU8);
   memcpy (SCM_BYTEVECTOR_CONTENTS (utf), c_utf, c_utf_len);
   free (c_utf);
@@ -2138,5 +2232,5 @@ scm_bootstrap_bytevectors (void)
 void
 scm_init_bytevectors (void)
 {
-#include "libguile/bytevectors.x"
+#include "bytevectors.x"
 }
