@@ -1,5 +1,5 @@
 ;;; Type analysis on CPS
-;;; Copyright (C) 2014-2019 Free Software Foundation, Inc.
+;;; Copyright (C) 2014-2021, 2023 Free Software Foundation, Inc.
 ;;;
 ;;; This library is free software: you can redistribute it and/or modify
 ;;; it under the terms of the GNU Lesser General Public License as
@@ -79,7 +79,6 @@
 (define-module (language cps types)
   #:use-module (ice-9 match)
   #:use-module (language cps)
-  #:use-module (language cps utils)
   #:use-module (language cps intmap)
   #:use-module (language cps intset)
   #:use-module (rnrs bytevectors)
@@ -127,6 +126,7 @@
             type<=?
 
             ;; Interface for type inference.
+            constant-type
             infer-types
             lookup-pre-type
             lookup-post-type
@@ -195,6 +195,8 @@
   (identifier-syntax (logior &fixnum &bignum &fraction)))
 (define-syntax &real
   (identifier-syntax (logior &fixnum &bignum &flonum &fraction)))
+(define-syntax &heap-number
+  (identifier-syntax (logior &flonum &bignum &complex &fraction)))
 (define-syntax &number
   (identifier-syntax (logior &fixnum &bignum &flonum &complex &fraction)))
 
@@ -342,8 +344,8 @@
 minimum, and maximum."
   (define (return type val)
     (if val
-        (make-type-entry type val val)
-        (make-type-entry type -inf.0 +inf.0)))
+        (values type val val)
+        (values type -inf.0 +inf.0)))
   (cond
    ((number? val)
     (cond
@@ -356,8 +358,8 @@ minimum, and maximum."
               val))
      ((eqv? (imag-part val) 0)
       (if (nan? val)
-          (make-type-entry &flonum -inf.0 +inf.0)
-          (make-type-entry
+          (values &flonum -inf.0 +inf.0)
+          (values
            (if (exact? val) &fraction &flonum)
            (if (rational? val) (inexact->exact (floor val)) val)
            (if (rational? val) (inexact->exact (ceiling val)) val))))
@@ -367,6 +369,7 @@ minimum, and maximum."
    ((eq? val #t) (return &special-immediate &true))
    ((eq? val #f) (return &special-immediate &false))
    ((eqv? val *unspecified*) (return &special-immediate &unspecified))
+   ((eof-object? val) (return &special-immediate &eof))
    ((char? val) (return &char (char->integer val)))
    ((symbol? val) (return &symbol #f))
    ((keyword? val) (return &keyword #f))
@@ -382,6 +385,13 @@ minimum, and maximum."
 
    (else (error "unhandled constant" val))))
 
+(define (constant-type-entry val)
+  "Compute the type and range of VAL.  Return three values: the type,
+minimum, and maximum."
+  (call-with-values (lambda () (constant-type val))
+    (lambda (type min max)
+      (make-type-entry type min max))))
+
 (define *type-checkers* (make-hash-table))
 (define *type-inferrers* (make-hash-table))
 
@@ -570,25 +580,28 @@ minimum, and maximum."
 ;;; Generic effect-free predicates.
 ;;;
 
-(define-syntax-rule (define-special-immediate-predicate-inferrer pred imm)
-  (define-predicate-inferrer (pred val true?)
+(define-syntax-rule (infer-constant-comparison ctype cval val true?)
+  (let ()
     (define (range-subtract lo hi x)
       (values (if (eqv? lo x) (1+ lo) lo)
               (if (eqv? hi x) (1- hi) hi)))
-    (cond
-     (true? (restrict! val &special-immediate imm imm))
-     (else
-      (when (eqv? (&type val) &special-immediate)
-        (let-values (((lo hi) (range-subtract (&min val) (&max val) imm)))
-          (restrict! val &special-immediate lo hi)))))))
+   (cond
+    (true? (restrict! val ctype cval cval))
+    (else
+     (when (eqv? (&type val) ctype)
+       (let-values (((lo hi) (range-subtract (&min val) (&max val) cval)))
+         (restrict! val ctype lo hi)))))))
 
-(define-special-immediate-predicate-inferrer eq-nil? &nil)
-(define-special-immediate-predicate-inferrer eq-eol? &null)
-(define-special-immediate-predicate-inferrer eq-false? &false)
-(define-special-immediate-predicate-inferrer eq-true? &true)
-(define-special-immediate-predicate-inferrer unspecified? &unspecified)
-(define-special-immediate-predicate-inferrer undefined? &undefined)
-(define-special-immediate-predicate-inferrer eof-object? &eof)
+(define-predicate-inferrer/param (eq-constant? c val true?)
+  (call-with-values (lambda () (constant-type c))
+    (lambda (ctype cval cval*)
+      ;; Either (= cval cval*), or the value is meaningless for this type.
+      (infer-constant-comparison ctype cval val true?))))
+
+;; Can't usefully pass undefined as a parameter to eq-constant?, so we
+;; keep its special predicate.
+(define-predicate-inferrer (undefined? val true?)
+  (infer-constant-comparison &special-immediate &undefined val true?))
 
 ;; Various inferrers rely on these having contiguous values starting from 0.
 (eval-when (expand)
@@ -622,13 +635,6 @@ minimum, and maximum."
     (logand &all-types (lognot &immediate-types)))
   (restrict! val (if true? &heap-object-types &immediate-types) -inf.0 +inf.0))
 
-(define-predicate-inferrer (heap-number? val true?)
-  (define &heap-number-types
-    (logior &bignum &flonum &complex &fraction))
-  (define &other-types
-    (logand &all-types (lognot &heap-number-types)))
-  (restrict! val (if true? &heap-number-types &other-types) -inf.0 +inf.0))
-
 (define-predicate-inferrer (fixnum? val true?)
   (cond
    (true?
@@ -663,27 +669,49 @@ minimum, and maximum."
 
 (define-syntax-rule (define-simple-predicate-inferrer predicate type)
   (define-predicate-inferrer (predicate val true?)
-    (let ((type (if true?
-                    type
-                    (logand (&type val) (lognot type)))))
-      (restrict! val type -inf.0 +inf.0))))
+    (restrict! val (if true? type (lognot type)) -inf.0 +inf.0)))
 
-(define-simple-predicate-inferrer pair? &pair)
-(define-simple-predicate-inferrer symbol? &symbol)
-(define-simple-predicate-inferrer variable? &box)
-(define-simple-predicate-inferrer immutable-vector? &immutable-vector)
-(define-simple-predicate-inferrer mutable-vector? &mutable-vector)
-(define-simple-predicate-inferrer struct? &struct)
-(define-simple-predicate-inferrer string? &string)
-(define-simple-predicate-inferrer bytevector? &bytevector)
+(define-simple-predicate-inferrer bignum? &bignum)
 (define-simple-predicate-inferrer bitvector? &bitvector)
-(define-simple-predicate-inferrer keyword? &keyword)
-(define-simple-predicate-inferrer number? &number)
+(define-simple-predicate-inferrer bytevector? &bytevector)
 (define-simple-predicate-inferrer char? &char)
-(define-simple-predicate-inferrer procedure? &procedure)
-(define-simple-predicate-inferrer flonum? &flonum)
 (define-simple-predicate-inferrer compnum? &complex)
+(define-simple-predicate-inferrer flonum? &flonum)
+(define-simple-predicate-inferrer fixnum? &fixnum)
+(define-simple-predicate-inferrer fluid? &fluid)
 (define-simple-predicate-inferrer fracnum? &fraction)
+(define-simple-predicate-inferrer immutable-vector? &immutable-vector)
+(define-simple-predicate-inferrer keyword? &keyword)
+(define-simple-predicate-inferrer mutable-vector? &mutable-vector)
+(define-simple-predicate-inferrer pair? &pair)
+(define-simple-predicate-inferrer pointer? &pointer)
+(define-simple-predicate-inferrer program? &procedure)
+(define-simple-predicate-inferrer string? &string)
+(define-simple-predicate-inferrer struct? &struct)
+(define-simple-predicate-inferrer symbol? &symbol)
+(define-simple-predicate-inferrer syntax? &syntax)
+(define-simple-predicate-inferrer variable? &box)
+
+(define-simple-predicate-inferrer number? &number)
+(define-type-inferrer-aliases number? rational? complex?)
+(define-simple-predicate-inferrer heap-number? &heap-number)
+(define-simple-predicate-inferrer real? &real)
+(let ((&maybe-integer (logior &exact-integer &flonum &complex)))
+  (define-simple-predicate-inferrer integer? &maybe-integer))
+(define-simple-predicate-inferrer exact-integer? &exact-integer)
+(define-simple-predicate-inferrer exact? &exact-number)
+(let ((&inexact-number (logior &flonum &complex)))
+  (define-simple-predicate-inferrer inexact? &inexact-number))
+
+(define-type-inferrer-aliases eq? heap-numbers-equal?)
+
+(define-predicate-inferrer (procedure? val true?)
+  ;; Besides proper procedures, structs and smobs can also be applicable
+  ;; in the guile-vm target.
+  (define applicable-types (logior &procedure &struct &other-heap-object))
+  (when true?
+    (restrict! val (logand (&type val) applicable-types)
+               (&min val) (&max val))))
 
 (define-predicate-inferrer (vector? val true?)
   (define &not-vector (logand &all-types (lognot &vector)))
@@ -702,7 +730,7 @@ minimum, and maximum."
 
 
 (define-type-inferrer/param (load-const/unlikely param result)
-  (let ((ent (constant-type param)))
+  (let ((ent (constant-type-entry param)))
     (define! result (type-entry-type ent)
       (type-entry-min ent) (type-entry-max ent))))
 
@@ -719,7 +747,80 @@ minimum, and maximum."
 
 
 ;;;
-;;; Memory.
+;;; High-level object representation.
+;;;
+
+(define-type-inferrer/param (allocate-vector param size result)
+  (define! result &vector (&min/0 size) (&max/scm-size size)))
+(define-type-inferrer/param (allocate-vector/immediate param result)
+  (define size param)
+  (define! result &vector size size))
+(define-type-inferrer (vector-length v result)
+  (define! result &u64 (&min/0 v) (&max/scm-size v)))
+(define-type-inferrer (vector-ref v idx result)
+  (restrict! v &vector (1+ (&min/0 idx)) (target-max-size-t/scm))
+  (define! result &all-types -inf.0 +inf.0))
+(define-type-inferrer/param (vector-ref/immediate param v result)
+  (define idx param)
+  (restrict! v &vector (1+ idx) (target-max-size-t/scm))
+  (define! result &all-types -inf.0 +inf.0))
+(define-type-inferrer (vector-set! v idx val)
+  (restrict! v &vector (1+ (&min/0 idx)) (target-max-size-t/scm)))
+(define-type-inferrer/param (vector-set!/immediate param v val)
+  (define idx param)
+  (restrict! v &vector (1+ param) (target-max-size-t/scm)))
+
+(define-type-inferrer (cons head tail result)
+  (define! result &pair -inf.0 +inf.0))
+(define-type-inferrer (box val result)
+  (define! result &box -inf.0 +inf.0))
+;; No inferrers for pair or box accessors; because type checks dominate
+;; these accessors, they would add no information.
+
+(define-type-inferrer/param (allocate-struct param vtable result)
+  (define nfields param)
+  ;; It would be nice to be able to restrict the vtable-size of vtable,
+  ;; but because vtables are themselves structs which have associated
+  ;; size ranges, there's nowhere to put the vtable-size ranges.  Humm!
+  (define! result &struct nfields nfields))
+(define-type-inferrer (vtable-size vtable result)
+  (define! result &u64 0 (target-max-size-t/scm)))
+;; No predicate inferrers for vtable-has-unboxed-fields? and
+;; vtable-field-boxed?, as there is nowhere to store this info.
+(define-type-inferrer (struct-vtable struct result)
+  (define! result &struct 0 (target-max-size-t/scm)))
+(define-type-inferrer/param (struct-ref param struct result)
+  (define idx param)
+  (restrict! struct &struct (1+ idx) (target-max-size-t/scm))
+  (define! result &all-types -inf.0 +inf.0))
+(define-type-inferrer/param (struct-set! param struct val)
+  (define idx param)
+  (restrict! struct &struct (1+ idx) (target-max-size-t/scm)))
+
+(define-type-inferrer (bv-contents bv result)
+  (define! result &other-heap-object -inf.0 +inf.0))
+(define-type-inferrer (bv-length bv result)
+  (define! result &u64 (&min/0 bv) (&max/size bv)))
+
+(define-type-inferrer (string-length str result)
+  (define! result &u64 (&min/0 str) (&max/size str)))
+(define-type-inferrer (string-ref str idx result)
+  (define! result &u64 0 *max-codepoint*))
+
+(define-type-inferrer (symbol-hash sym result)
+  (define! result &u64 0 &u64-max))
+
+(define-type-inferrer/param (make-closure param code result)
+  (define nfree param)
+  (define! result &procedure nfree nfree))
+;; No information would be provided by closure-ref / closure-set!
+;; inferrers.
+
+
+
+
+;;;
+;;; Low-level object representation.
 ;;;
 
 (define (annotation->type ann)
@@ -728,11 +829,13 @@ minimum, and maximum."
     ('vector &vector)
     ('string &string)
     ('stringbuf &string)
+    ('symbol &symbol)
     ('bytevector &bytevector)
     ('box &box)
     ('closure &procedure)
     ('struct &struct)
-    ('atomic-box &all-types)))
+    ('atomic-box &all-types)
+    ('keyword &keyword)))
 
 (define (annotation->mutable-type ann)
   (match ann
@@ -831,7 +934,21 @@ minimum, and maximum."
 
 
 ;;;
-;;; Threads.  We don't currently track threads as an object type.
+;;; Symbols and keywords
+;;;
+(define-simple-types
+  ((symbol->keyword &symbol) &keyword)
+  ((keyword->symbol &keyword) &symbol)
+  ((symbol->string &symbol) &string)
+  ((string->symbol &string) &symbol)
+  ((string-utf8-length &string) &u64)
+  ((utf8->string &bytevector) &string))
+
+
+
+
+;;;
+;;;  We don't currently track threads as an object type.
 ;;;
 
 (define-simple-types
@@ -1099,7 +1216,7 @@ minimum, and maximum."
                          (+ (&min a) (&min b))
                          (+ (&max a) (&max b))))
 (define-type-inferrer/param (add/immediate param a result)
-  (let ((b-type (type-entry-type (constant-type param))))
+  (let ((b-type (type-entry-type (constant-type-entry param))))
     (define-binary-result! (&type a) b-type result #t
       (+ (&min a) param)
       (+ (&max a) param))))
@@ -1143,7 +1260,7 @@ minimum, and maximum."
                          (- (&min a) (&max b))
                          (- (&max a) (&min b))))
 (define-type-inferrer/param (sub/immediate param a result)
-  (let ((b-type (type-entry-type (constant-type param))))
+  (let ((b-type (type-entry-type (constant-type-entry param))))
     (define-binary-result! (&type a) b-type result #t
       (- (&min a) param)
       (- (&max a) param))))
@@ -1332,16 +1449,6 @@ minimum, and maximum."
    (else
     (define! result &special-immediate &false &true))))
 
-(define-simple-type-checker (exact? &number))
-(define-type-inferrer (exact? val result)
-  (restrict! val &number -inf.0 +inf.0)
-  (define-type-predicate-result val result &exact-number))
-
-(define-simple-type-checker (inexact? &number))
-(define-type-inferrer (inexact? val result)
-  (restrict! val &number -inf.0 +inf.0)
-  (define-type-predicate-result val result (logior &flonum &complex)))
-
 (define-simple-type-checker (inf? &real))
 (define-type-inferrer (inf? val result)
   (restrict! val &real -inf.0 +inf.0)
@@ -1383,6 +1490,7 @@ minimum, and maximum."
                                              (&min/0 count)
                                              (&max/u64 count))))
     (define-exact-integer! result min max)))
+(define-simple-type-checker (lsh/immediate &exact-integer))
 (define-type-inferrer/param (lsh/immediate count val result)
   (restrict! val &exact-integer -inf.0 +inf.0)
   (let-values (((min max) (compute-ash-range (&min val)
@@ -1396,6 +1504,7 @@ minimum, and maximum."
                                              (- (&min/0 count))
                                              (- (&max/u64 count)))))
     (define-exact-integer! result min max)))
+(define-simple-type-checker (rsh/immediate &exact-integer))
 (define-type-inferrer/param (rsh/immediate count val result)
   (restrict! val &exact-integer -inf.0 +inf.0)
   (let-values (((min max) (compute-ash-range (&min val)
@@ -1429,7 +1538,9 @@ minimum, and maximum."
         (define! result &s64 &s64-min &s64-max))))
 
 (define-type-inferrer (ulsh a b result)
-  (if (<= (ash (&max/u64 a) (&max/u64 b)) &u64-max)
+  (if (and
+       (or (zero? (&max/u64 a)) (< (&max/u64 b) 64)) ; don't even try
+       (<= (ash (&max/u64 a) (&max/u64 b)) &u64-max))
       ;; No overflow; we can be precise.
       (define! result &u64
         (ash (&min/0 a) (&min/0 b))
@@ -1533,10 +1644,24 @@ where (A0 <= A <= A1) and (B0 <= B <= B1)."
     (lambda (min max)
       (define-exact-integer! result min max))))
 
+(define-simple-type-checker (logand/immediate &exact-integer))
+(define-type-inferrer/param (logand/immediate param a result)
+  (restrict! a &exact-integer -inf.0 +inf.0)
+  (call-with-values (lambda ()
+                      (logand-bounds (&min a) (&max a) param param))
+    (lambda (min max)
+      (define-exact-integer! result min max))))
+
 (define-type-inferrer (ulogand a b result)
   (restrict! a &u64 0 &u64-max)
   (restrict! b &u64 0 &u64-max)
   (define! result &u64 0 (min (&max/u64 a) (&max/u64 b))))
+(define-type-inferrer/param (ulogand/immediate param a result)
+  (restrict! a &u64 0 &u64-max)
+  (call-with-values (lambda ()
+                      (logand-bounds (&min a) (&max a) param param))
+    (lambda (min max)
+      (define! result &u64 min max))))
 
 (define (logsub-bounds a0 a1 b0 b1)
   "Return two values: lower and upper bounds for (logsub A B),
@@ -1857,6 +1982,7 @@ where (A0 <= A <= A1) and (B0 <= B <= B1)."
     (($ $kargs _ _ ($ $throw)) 0)
     (($ $kargs _ _ ($ $continue)) 1)
     (($ $kargs _ _ (or ($ $branch) ($ $prompt))) 2)
+    (($ $kargs _ _ ($ $switch kf kt*)) (1+ (length kt*)))
     (($ $kfun src meta self tail clause) (if clause 1 0))
     (($ $kclause arity body alt) (if alt 2 1))
     (($ $kreceive) 1)
@@ -1998,7 +2124,7 @@ maximum, where type is a bitset as a fixnum."
                    (match (intmap-ref conts k)
                      (($ $kargs _ defs)
                       (infer-primcall types 0 name param args
-                                      (match defs ((var) var) (() #f))))
+                                      (match defs ((var) var) (_ #f))))
                      (_
                       ;; (pk 'warning-no-restrictions name)
                       types))))
@@ -2015,7 +2141,7 @@ maximum, where type is a bitset as a fixnum."
                      (adjoin-var out def (var-type-entry in arg))))))))
          (_
           (propagate1 k types))))
-      ((or ($ $call) ($ $callk))
+      ((or ($ $call) ($ $callk) ($ $calli))
        (propagate1 k types))
       (($ $rec names vars funs)
        (let ((proc-type (make-type-entry &procedure -inf.0 +inf.0)))
@@ -2025,7 +2151,7 @@ maximum, where type is a bitset as a fixnum."
          (($ $kargs (_) (var))
           (let ((entry (match exp
                          (($ $const val)
-                          (constant-type val))
+                          (constant-type-entry val))
                          ((or ($ $prim) ($ $fun) ($ $const-fun) ($ $code))
                           ;; Could be more precise here.
                           (make-type-entry &procedure -inf.0 +inf.0)))))
@@ -2053,6 +2179,24 @@ maximum, where type is a bitset as a fixnum."
          ;; The "normal" continuation is the #f branch.
          (propagate2 kf (infer-primcall types 0 op param args #f)
                      kt (infer-primcall types 1 op param args #f)))
+        (($ $kargs names vars ($ $switch kf kt* src arg))
+         (define (restrict-index min max)
+           (restrict-var types arg (make-type-entry &u64 min max)))
+         (define (visit-default typev)
+           (let ((types (restrict-index (length kt*) &u64-max)))
+             (propagate-types label typev 0 kf types)))
+         (define (visit-target typev k i)
+           (let ((types (restrict-index i i)))
+             (propagate-types label typev (1+ i) k types)))
+         (call-with-values (lambda () (visit-default typev))
+           (lambda (changed typev)
+             (let lp ((kt* kt*) (i 0) (changed changed) (typev typev))
+               (match kt*
+                 (() (values changed typev))
+                 ((kt . kt*)
+                  (call-with-values (lambda () (visit-target typev kt i))
+                    (lambda (changed* typev)
+                      (lp kt* (1+ i) (append changed* changed) typev)))))))))
         (($ $kargs names vars ($ $prompt k kh src escape? tag))
          ;; The "normal" continuation enters the prompt.
          (propagate2 k types kh types))
@@ -2064,9 +2208,14 @@ maximum, where type is a bitset as a fixnum."
             (propagate1 k (adjoin-vars types vars all-types-entry)))))
         (($ $kfun src meta self tail clause)
          (if clause
-             (propagate1 clause (if self
-                                    (adjoin-var types self all-types-entry)
-                                    types))
+             (let ((types (if self
+                              (adjoin-var types self all-types-entry)
+                              types)))
+               (propagate1 clause
+                           (match (intmap-ref conts clause)
+                             (($ $kargs _ defs)
+                              (adjoin-vars types defs all-types-entry))
+                             (_ types))))
              (propagate0)))
         (($ $kclause arity kbody kalt)
          (match (intmap-ref conts kbody)
