@@ -1,6 +1,6 @@
 ;;; HTTP messages
 
-;; Copyright (C) 2010-2017, 2019 Free Software Foundation, Inc.
+;; Copyright (C)  2010-2017, 2023 Free Software Foundation, Inc.
 
 ;; This library is free software; you can redistribute it and/or
 ;; modify it under the terms of the GNU Lesser General Public
@@ -30,15 +30,13 @@
 ;;; Code:
 
 (define-module (web http)
-  #:use-module ((srfi srfi-1) #:select (append-map! map!))
   #:use-module (srfi srfi-9)
   #:use-module (srfi srfi-19)
   #:use-module (ice-9 rdelim)
   #:use-module (ice-9 match)
-  #:use-module (ice-9 q)
   #:use-module (ice-9 binary-ports)
   #:use-module (ice-9 textual-ports)
-  #:use-module (rnrs bytevectors)
+  #:use-module (ice-9 exceptions)
   #:use-module (web uri)
   #:export (string->header
             header->string
@@ -67,6 +65,8 @@
             read-response-line
             write-response-line
 
+            &chunked-input-error-prematurely
+            chunked-input-ended-prematurely-error?
             make-chunked-input-port
             make-chunked-output-port
 
@@ -152,35 +152,18 @@ The default writer will call ‘put-string’."
         (lambda (val port)
           (put-string port val)))))
 
-(define spaces-and-tabs
-  (char-set #\space #\tab))
-
-(define (space-or-tab? c)
-  (case c
-    ((#\space #\tab) #t)
-    (else #f)))
-
 (define (read-header-line port)
-  "Read an HTTP header line, including any continuation lines, and
-return the combined string without its final CRLF or LF.  Raise a
-'bad-header' exception if the line does not end in CRLF or LF, or if EOF
-is reached."
+  "Read an HTTP header line and return it without its final CRLF or LF.
+Raise a 'bad-header' exception if the line does not end in CRLF or LF,
+or if EOF is reached."
   (match (%read-line port)
     (((? string? line) . #\newline)
      ;; '%read-line' does not consider #\return a delimiter; so if it's
      ;; there, remove it.  We are more tolerant than the RFC in that we
      ;; tolerate LF-only endings.
-     (let ((line (if (string-suffix? "\r" line)
-                     (string-drop-right line 1)
-                     line)))
-       ;; If the next character is a space or tab, then there's at least
-       ;; one continuation line.  Read the continuation lines by calling
-       ;; 'read-header-line' recursively, and append them to this header
-       ;; line, folding the leading spaces and tabs to a single space.
-       (if (space-or-tab? (lookahead-char port))
-           (string-append line " " (string-trim (read-header-line port)
-                                                spaces-and-tabs))
-           line)))
+     (if (string-suffix? "\r" line)
+         (string-drop-right line 1)
+         line))
     ((line . _)                                ;EOF or missing delimiter
      (bad-header 'read-header-line line))))
 
@@ -979,13 +962,23 @@ as an ordered alist."
     (((? symbol?) . (? key-value-list?)) #t)
     (_ #f)))
 
+;; While according to RFC 7617 Schemes are case-insensitive:
+;;
+;; 'Note that both scheme and parameter names are matched
+;; case-insensitive'
+;;
+;; some software (*) incorrectly assumes title case for scheme
+;; names, so use the more titlecase.
+;;
+;; (*): See, e.g.,
+;; https://community.spotify.com/t5/Spotify-for-Developers/API-Authorization-header-doesn-t-follow-HTTP-spec/m-p/5397381#M4917
 (define (write-credentials val port)
   (match val
     (('basic . cred)
-     (put-string port "basic ")
+     (put-string port "Basic ")
      (put-string port cred))
     ((scheme . params)
-     (put-symbol port scheme)
+     (put-string port (string-titlecase (symbol->string scheme)))
      (put-char port #\space)
      (write-key-value-list params port))))
 
@@ -1125,6 +1118,8 @@ symbol, like ‘GET’."
    ((string= str "DELETE" start end) 'DELETE)
    ((string= str "OPTIONS" start end) 'OPTIONS)
    ((string= str "TRACE" start end) 'TRACE)
+   ((string= str "CONNECT" start end) 'CONNECT)
+   ((string= str "PATCH" start end) 'PATCH)
    (else (bad-request "Invalid method: ~a" (substring str start end)))))
 
 (define* (parse-request-uri str #:optional (start 0) (end (string-length str)))
@@ -1950,6 +1945,17 @@ treated specially, and is just returned as a plain string."
 
 
 ;; Chunked Responses
+(define &chunked-input-ended-prematurely
+  (make-exception-type '&chunked-input-error-prematurely
+                       &external-error
+                       '()))
+
+(define make-chunked-input-ended-prematurely-error
+  (record-constructor &chunked-input-ended-prematurely))
+
+(define chunked-input-ended-prematurely-error?
+  (record-predicate &chunked-input-ended-prematurely))
+
 (define (read-chunk-header port)
   "Read a chunk header from PORT and return the size in bytes of the
 upcoming chunk."
@@ -1992,6 +1998,7 @@ closed it will also close PORT, unless the KEEP-ALIVE? is true."
                (cond
                 ((zero? size)
                  (set! finished? #t)
+                 (get-bytevector-n port 2) ; \r\n follows the last chunk
                  num-read)
                 (else
                  (loop to-read num-read)))))
@@ -2001,8 +2008,8 @@ closed it will also close PORT, unless the KEEP-ALIVE? is true."
                                                 ask-for)))
                (cond
                 ((eof-object? read)     ;premature termination
-                 (set! finished? #t)
-                 num-read)
+                 (raise-exception
+                  (make-chunked-input-ended-prematurely-error)))
                 (else
                  (let ((left (- remaining read)))
                    (set! remaining left)
@@ -2025,34 +2032,23 @@ BUFFERING bytes, which defaults to 1200.  Take care to close the port
 when done, as it will output the remaining data, and encode the final
 zero chunk. When the port is closed it will also close PORT, unless
 KEEP-ALIVE? is true."
-  (define (q-for-each f q)
-    (while (not (q-empty? q))
-      (f (deq! q))))
-  (define queue (make-q))
-  (define (%put-char c)
-    (enq! queue c))
-  (define (%put-string s)
-    (string-for-each (lambda (c) (enq! queue c))
-                     s))
-  (define (flush)
-    ;; It is important that we do _not_ write a chunk if the queue is
-    ;; empty, since it will be treated as the final chunk.
-    (unless (q-empty? queue)
-      (let ((len (q-length queue)))
-        (put-string port (number->string len 16))
-        (put-string port "\r\n")
-        (q-for-each (lambda (elem) (put-char port elem))
-                    queue)
-        (put-string port "\r\n"))))
+  (define (write! bv start count)
+    (put-string port (number->string count 16))
+    (put-string port "\r\n")
+    (put-bytevector port bv start count)
+    (put-string port "\r\n")
+    (force-output port)
+    count)
   (define (close)
-    (flush)
-    (put-string port "0\r\n")
+    (put-string port "0\r\n\r\n")
     (force-output port)
     (unless keep-alive?
       (close-port port)))
-  (let ((ret (make-soft-port (vector %put-char %put-string flush #f close) "w")))
-    (setvbuf ret 'block buffering)
-    ret))
+  (define ret
+    (make-custom-binary-output-port "chunked http" write! #f #f close))
+  (set-port-encoding! port "UTF-8")
+  (setvbuf ret 'block buffering)
+  ret)
 
 (define %http-proxy-port? (make-object-property))
 (define (http-proxy-port? port) (%http-proxy-port? port))
